@@ -1,43 +1,40 @@
 # -*- ispell-local-dictionary: "american" -*-
-# SPDX-License-Identifier: Python-2.0-like
+# SPDX-License-Identifier: PSF-2.0
 """ManageSieve (RFC 5804) client module for remotely managing Sieve Scripts.
 
 
 All Sieve commands are supported by methods of the same name (in
 lower-case). All arguments to commands are converted to strings,
-except for :meth:`authenticate`.
+except for :code:`authenticate`.
 
 """
 
-__version__ = "0.7.1"
+__version__ = "0.8"
 __author__ = """Hartmut Goebel <h.goebel@crazy-compilers.com>
 Ulrich Eck <ueck@net-labs.de> April 2001
 """
 
-__copyright__ = "Copyright (C) 2003-2021 by Hartmut Goebel <h.goebel@crazy-compilers.com> and others"
+__copyright__ = ("Copyright (C) 2003-2024 by Hartmut Goebel "
+                 "<h.goebel@crazy-compilers.com> and others")
 __license__ = "Python-2.0 like"
 
 import binascii
 import logging
 import re
 import socket
-from logging import log
-try:
-    import ssl
-    ssl_wrap_socket = ssl.wrap_socket
-except ImportError:
-    ssl_wrap_socket = socket.ssl
+import ssl
 
+__all__ = ['MANAGESIEVE', 'SIEVE_PORT', 'OK', 'NO', 'BYE',
+           'INFO', 'DEBUG', 'DEBUG0', 'DEBUG1', 'DEBUG2', 'DEBUG3']
 
-__all__ = [ 'MANAGESIEVE', 'SIEVE_PORT', 'OK', 'NO', 'BYE',
-            'INFO', 'DEBUG', 'DEBUG0', 'DEBUG1', 'DEBUG2', 'DEBUG3']
+logger = logging.getLogger('managesieve')
 
 # logging levels
 INFO = logging.INFO
-DEBUG0 = DEBUG = logging.DEBUG+3 # commands and responses
-DEBUG1 = logging.DEBUG+2 # send and read data (except long literals)
-DEBUG2 = logging.DEBUG+1 # more details
-DEBUG3 = logging.DEBUG   # all debug messages (pattern matching, etc.)
+DEBUG0 = logging.DEBUG+1  # commands and responses
+DEBUG1 = DEBUG = logging.DEBUG  # send and read data (except long literals)
+DEBUG2 = logging.DEBUG-1  # more details
+DEBUG3 = logging.DEBUG-2  # all debug messages (pattern matching, etc.)
 
 CRLF = b'\r\n'
 SIEVE_PORT = 4190
@@ -53,33 +50,34 @@ AUTH_LOGIN = "LOGIN"
 AUTHMECHS = [AUTH_PLAIN, AUTH_LOGIN]
 
 # todo: return results or raise exceptions?
-# todo: on result 'BYE' quit immediatly
+# todo: on result 'BYE' quit immediately
 # todo: raise exception on 'BYE'?
 
 #    Commands
 commands = {
     # name            valid states
-    b'STARTTLS':     ('NONAUTH',),
-    b'AUTHENTICATE': ('NONAUTH',),
-    b'LOGOUT':       ('NONAUTH', 'AUTH', 'LOGOUT'),
-    b'CAPABILITY':   ('NONAUTH', 'AUTH'),
-    b'GETSCRIPT':    ('AUTH', ),
-    b'PUTSCRIPT':    ('AUTH', ),
-    b'SETACTIVE':    ('AUTH', ),
-    b'DELETESCRIPT': ('AUTH', ),
-    b'LISTSCRIPTS':  ('AUTH', ),
-    b'HAVESPACE':    ('AUTH', ),
+    b'STARTTLS':       ('NONAUTH',),
+    b'AUTHENTICATE':   ('NONAUTH',),
+    b'UNAUTHENTICATE': ('AUTH',),
+    b'LOGOUT':         ('NONAUTH', 'AUTH', 'LOGOUT'),
+    b'CAPABILITY':     ('NONAUTH', 'AUTH'),
+    b'GETSCRIPT':      ('AUTH', ),
+    b'PUTSCRIPT':      ('AUTH', ),
+    b'SETACTIVE':      ('AUTH', ),
+    b'DELETESCRIPT':   ('AUTH', ),
+    b'LISTSCRIPTS':    ('AUTH', ),
+    b'HAVESPACE':      ('AUTH', ),
     }
 
-### needed
-Oknobye = re.compile(r'(?P<type>(?:OK|NO|BYE))' # atom / ascii
-                     r'(?: \((?P<code>.*)\))?'  # atom / ascii
-                     r'(?: (?P<data>.*))?')     # string / utf-8
+# -- needed
+Oknobye = re.compile(r'(?P<type>(?:OK|NO|BYE))'  # atom / ascii
+                     r'(?: \((?P<code>.*)\))?'   # atom / ascii
+                     r'(?: (?P<data>.*))?')      # string / utf-8
 # draft-martin-managesieve-04.txt defines the size tag of literals to
 # contain a '+' (plus sign) behind the digits, but timsieved does not
 # send one. Thus we are less strikt here:
 Literal = re.compile(r'.*{(?P<size>\d+)\+?}$')
-re_dquote  = re.compile(r'"(([^"\\]|\\.)*)"')
+re_dquote = re.compile(r'"(([^"\\]|\\.)*)"')
 re_esc_quote = re.compile(r'\\([\\"])')
 
 
@@ -98,33 +96,14 @@ class SSLFakeSocket:
 
     sendall = send
 
-    def close(self):
-        self.realsock.close()
-
-class SSLFakeFile:
-    """A fake file like object that really wraps a SSLObject.
-
-    It only supports what is needed in managesieve.
-    """
-    def __init__(self, sslobj):
-        self.sslobj = sslobj
-
-    def readline(self):
-        str = b""
-        chr = None
-        while chr != b"\n":
-            chr = self.sslobj.read(1)
-            str += chr
-        return str
-
-    def read(self, size=0):
-        if size == 0:
+    def recv(self, bufsize):
+        if bufsize == 0:
             return b''
         else:
-            return self.sslobj.read(size)
+            return self.sslobj.read(bufsize)
 
     def close(self):
-        pass
+        self.realsock.close()
 
 
 def sieve_name(name):
@@ -156,8 +135,10 @@ class MANAGESIEVE:
     :param port: port number (default: standard Sieve port).
     :param use_tls:  switch to TLS automatically,
                      fail if the server doesn't support STARTTLS
-    :param keyfile:  keyfile to use for TLS (optional)
-    :param certfile: certfile to use for TLS (optional)
+    :param keyfile:  keyfile to use for TLS client authentication (optional)
+    :param certfile: certfile to use for TLS client authentication (optional)
+    :param timeout:  If not None, socket timeout in seconds (refer to python's
+                     `socket.settimeout` for details)
     """
 
     """
@@ -183,8 +164,8 @@ class MANAGESIEVE:
     the results.
     """
 
-    class error(Exception): """Logical errors - debug required"""
-    class abort(error):     """Service errors - close and retry"""
+    class error(Exception): """Logical errors - debug required"""   # noqa:E701
+    class abort(error):     """Service errors - close and retry"""  # noqa:E701
 
     def __clear_knowledge(self):
         """clear/init any knowledge obtained from the server"""
@@ -194,22 +175,24 @@ class MANAGESIEVE:
         self.supports_tls = 0
 
     def __init__(self, host='', port=SIEVE_PORT,
-                 use_tls=False, keyfile=None, certfile=None):
+                 use_tls=False, keyfile=None, certfile=None, tls_verify=True,
+                 timeout=None):
         self.host = host
         self.port = port
         self.state = 'NONAUTH'
+        self.timeout = timeout
 
         self.response_text = self.response_code = None
         self.__clear_knowledge()
 
-        # Open socket to server.
-        self._open(host, port)
-
         if __debug__:
+            self._log(INFO, 'managesieve version %s', __version__)
             self._cmd_log_len = 10
             self._cmd_log_idx = 0
             self._cmd_log = {}           # Last `_cmd_log_len' interactions
-            self._log(INFO, 'managesieve version %s', __version__)
+
+        # Open socket to server.
+        self._open(host, port)
 
         # Get server welcome message,
         # request and store CAPABILITY response.
@@ -219,11 +202,11 @@ class MANAGESIEVE:
         if use_tls:
             if not self.supports_tls:
                 self.abort('TLS requested, but server does not support TLS')
-            typ, data = self.starttls(keyfile=keyfile, certfile=certfile)
+            typ, data = self.starttls(keyfile=keyfile, certfile=certfile,
+                                      verify=tls_verify)
             if typ == 'OK':
                 self._parse_capabilities(data)
 
-
     def _parse_capabilities(self, lines):
         for line in lines:
             if len(line) == 2:
@@ -248,37 +231,41 @@ class MANAGESIEVE:
                 pass
         return
 
-
     def __getattr__(self, attr):
         #    Allow UPPERCASE variants of MANAGESIEVE command methods.
         if attr in commands:
             return getattr(self, attr.lower())
         raise AttributeError("Unknown MANAGESIEVE command: '%s'" % attr)
 
-
-    #### Private methods ###
+    # ### Private methods ###
     def _open(self, host, port):
-        """Setup 'self.sock' and 'self.file'."""
-        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        self.sock.connect((self.host, self.port))
-        self.file = self.sock.makefile('rb')
+        """Setup 'self.sock'."""
+        self.sock = socket.create_connection(
+            (self.host, self.port), timeout=self.timeout)
+        if __debug__:
+            self._log(INFO, 'Connected to %s', self.sock.getpeername()[0])
 
     def _close(self):
-        self.file.close()
         self.sock.close()
 
     def _read(self, size):
         """Read 'size' bytes from remote."""
         data = b""
         while len(data) < size:
-            data += self.file.read(size - len(data))
+            # NOTE: Before python 3.5 this could raise an InterruptedError (and
+            # would have to be retried in such a case)
+            data += self.sock.recv(size - len(data))
         return data
 
     def _readline(self):
-        """Read line from remote."""
-        # Continuation bytes are never 7-bit-ascii characters, so
-        # using readline() scanning for CR and LR is okay here.
-        line = self.file.readline()
+        """Read one line from remote."""
+        line = b''
+        last = (None, None)
+        # Protocol mandates all lines terminated by CRLF
+        while last != (b'\r', b'\n'):
+            chr = self.sock.recv(1)
+            last = (last[1], chr)
+            line += chr
         assert isinstance(line, bytes)
         line = line.decode('utf-8')
         assert isinstance(line, str)
@@ -289,8 +276,6 @@ class MANAGESIEVE:
 
     def _get_line(self):
         line = self._readline()
-        if not line:
-            raise self.abort('socket error: EOF')
         assert isinstance(line, str)
         # Protocol mandates all lines terminated by CRLF
         line = line[:-2]
@@ -305,19 +290,18 @@ class MANAGESIEVE:
            typ  = response type
 
         The response code and text may be found in :attr:`response_code`
-        and :attr:`response_text`, respectivly.
+        and :attr:`response_text`, respectively.
         """
-        return self._command(*args)[0] # only return typ, ignore data
-
+        return self._command(*args)[0]  # only return typ, ignore data
 
     def _command(self, name, arg1=None, arg2=None, *options):
         """
         Returns (typ, data) with
            typ  = response type
-           data = list of lists of strings read (only meaningfull if OK)
+           data = list of lists of strings read (only meaningful if OK)
 
         The response code and text may be found in :attr:`.response_code`
-        and :attr:`response_text`, respectivly.
+        and :attr:`response_text`, respectively.
         """
         assert isinstance(name, bytes)
         if self.state not in commands[name]:
@@ -342,12 +326,11 @@ class MANAGESIEVE:
                 self._print_log()
             raise
 
-
     def _readstring(self, data):
         assert isinstance(data, str)
-        if data.startswith(" ") : # space -> error
+        if data.startswith(" "):  # space -> error
             raise self.error('Unexpected space: %r' % data)
-        elif data.startswith('"'): # handle double quote:
+        elif data.startswith('"'):  # handle double quote:
             if not self._match(re_dquote, data):
                 raise self.error('Unmatched quote: %r' % data)
             snippet = self.mo.group(1)
@@ -369,10 +352,10 @@ class MANAGESIEVE:
         :returns: :tuple:(resp, data) with
 
            :resp: response (OK, NO, BYE)
-           :data: list of lists of strings read (only meaningfull if OK)
+           :data: list of lists of strings read (only meaningful if OK)
 
         The response code and text may be found in :attr:`response_code`
-        and :attr:`response_text`, respectivly.
+        and :attr:`response_text`, respectively.
         """
 
         """
@@ -380,7 +363,8 @@ class MANAGESIEVE:
     response-authenticate = *(string CRLF) (response-oknobye)
     response-capability   = *(string [SP string] CRLF) response-oknobye
     response-listscripts  = *(string [SP "ACTIVE"] CRLF) response-oknobye
-    response-oknobye      = ("OK" / "NO" / "BYE") [SP "(" resp-code ")"] [SP string] CRLF
+    response-oknobye      = ("OK" / "NO" / "BYE") [SP "(" resp-code ")"]
+                            [SP string] CRLF
     string                = quoted / literal
     quoted                = <"> *QUOTED-CHAR <">
     literal               = "{" number  "+}" CRLF *OCTET
@@ -397,12 +381,13 @@ class MANAGESIEVE:
 [A-Z-]+ CRLF
 
         """
-        data = [] ; dat = None
+        data = []
+        dat = None
         resp = self._get_line()
         assert isinstance(resp, str)
         while 1:
             if self._match(Oknobye, resp):
-                typ, code, dat = self.mo.group('type','code','data')
+                typ, code, dat = self.mo.group('type', 'code', 'data')
                 if __debug__:
                     self._log(DEBUG0, '%s response: %s %s', typ, code, dat)
                 self.response_code = code
@@ -431,58 +416,58 @@ class MANAGESIEVE:
                 resp = self._get_line()
         raise self.error('Should not come here')
 
-
     def _match(self, cre, s):
         # Run compiled regular expression match method on 's'.
         # Save result, return success.
         self.mo = cre.match(s)
         if __debug__ and self.mo is not None:
-            self._log(DEBUG3, "\tmatched '%r' => %r", cre.pattern, self.mo.groups())
+            self._log(DEBUG3, "\tmatched '%r' => %r",
+                      cre.pattern, self.mo.groups())
         return self.mo is not None
 
-
     if __debug__:
-
         def _log(self, level, msg, *args):
             if level == DEBUG1:
                 # Keep log of last `_cmd_log_len' interactions for debugging.
                 self._cmd_log[self._cmd_log_idx] = (msg, args)
                 self._cmd_log_idx = (self._cmd_log_idx+1) % self._cmd_log_len
-            log(level, msg, *args)
-            
+            logger.log(level, msg, *args)
+
         def _print_log(self):
             idx, cnt = self._cmd_log_idx, len(self._cmd_log)
-            log(logging.ERROR, 'last %d SIEVE interactions:', cnt)
+            logger.log(logging.ERROR, 'last %d SIEVE interactions:', cnt)
             idx = idx % cnt
             while cnt:
                 msg, args = self._cmd_log[idx]
                 try:
-                    log(logging.ERROR, msg, *args)
-                except:
+                    logger.log(logging.ERROR, msg, *args)
+                except Exception:
                     pass
                 idx = (idx+1) % self._cmd_log_len
                 cnt -= 1
 
-    ### Public methods ###
+    # ## Public methods ###
     def authenticate(self, mechanism, *authobjects):
         """Authenticate to the server.
 
         :param str mechanism: authentication mechanism to use
         :param authobjects: authentication data for this mechanism
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
-        # command-authenticate  = "AUTHENTICATE" SP auth-type [SP string]  *(CRLF string)
+        # command-authenticate  = "AUTHENTICATE" SP auth-type [SP string]
+        #                         *(CRLF string)
         # response-authenticate = *(string CRLF) (response-oknobye)
         mech = mechanism.upper()
-        if not mech in self.loginmechs:
+        if mech not in self.loginmechs:
             raise self.error("Server doesn't allow %s authentication." % mech)
 
         authobjects = [ao.encode('utf-8') for ao in authobjects]
         if mech == AUTH_LOGIN:
-            authobjects = [ sieve_name(binascii.b2a_base64(ao)[:-1].decode('ascii'))
-                            for ao in authobjects
-                            ]
+            authobjects = [
+                sieve_name(binascii.b2a_base64(ao)[:-1].decode('ascii'))
+                for ao in authobjects
+            ]
         elif mech == AUTH_PLAIN:
             if len(authobjects) < 3:
                 # assume authorization identity (authzid) is missing
@@ -490,9 +475,10 @@ class MANAGESIEVE:
                 authobjects.insert(0, b'')
             ao = b'\0'.join(authobjects)
             ao = binascii.b2a_base64(ao)[:-1].decode('ascii')
-            authobjects = [ sieve_string(ao) ]
+            authobjects = [sieve_string(ao)]
         else:
-            raise self.error("managesieve doesn't support %s authentication." % mech)
+            raise self.error("managesieve doesn't support %s authentication."
+                             % mech)
 
         typ, data = self._command(b'AUTHENTICATE',
                                   sieve_name(mech), *authobjects)
@@ -500,12 +486,23 @@ class MANAGESIEVE:
             self.state = 'AUTH'
         return typ
 
+    def unauthenticate(self):
+        """Leave the authenticated state.
+
+        :returns: response (``OK``, ``NO``, ``BYE``)
+        """
+        # command-unauthenticate     = "UNAUTHENTICATE" CRLF
+        # response-unauthenticate    = response-oknobye
+        typ, data = self._command(b'UNAUTHENTICATE')
+        if typ == 'OK':
+            self.state = 'NONAUTH'
+        return typ
 
     def login(self, auth, user, password):
         """
         Authenticate to the Sieve server using the best mechanism available.
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         for authmech in AUTHMECHS:
             if authmech in self.loginmechs:
@@ -519,7 +516,7 @@ class MANAGESIEVE:
     def logout(self):
         """Terminate connection to server.
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         # command-logout        = "LOGOUT" CRLF
         # response-logout       = response-oknobye
@@ -528,46 +525,46 @@ class MANAGESIEVE:
         self._close()
         return typ
 
-
     def listscripts(self):
         """Get a list of scripts on the server.
 
         :returns: tuple(response, [data]) --
-                  if `response` is :const:`OK`, `data` is a list of
+                  if `response` is ``OK``, `data` is a list of
                   `(scriptname, active)` tuples.
         """
         # command-listscripts   = "LISTSCRIPTS" CRLF
-        # response-listscripts  = *(sieve-name [SP "ACTIVE"] CRLF) response-oknobye
+        # response-listscripts  = *(sieve-name [SP "ACTIVE"] CRLF)
+        #                         response-oknobye
         typ, data = self._command(b'LISTSCRIPTS')
-        if typ != 'OK': return typ, data
+        if typ != 'OK':
+            return typ, data
         scripts = []
         for dat in data:
             if __debug__:
                 if not len(dat) in (1, 2):
                     raise self.error("Unexpected result from LISTSCRIPTS: %r"
                                      % (dat,))
-            scripts.append( (dat[0], dat[1] is not None ))
+            scripts.append((dat[0], dat[1] is not None))
         return typ, scripts
 
-
     def getscript(self, scriptname):
         """Get a script from the server.
 
         :param str scriptname: name of script to be retrieved
 
         :returns: tuple(response, str) --
-                  if `response` is :const:`OK`, `str` is the script content.
+                  if `response` is ``OK``, `str` is the script content.
         """
         # command-getscript     = "GETSCRIPT" SP sieve-name CRLF
         # response-getscript    = [string CRLF] response-oknobye
-        
+
         typ, data = self._command(b'GETSCRIPT', sieve_name(scriptname))
-        if typ != 'OK': return typ, data
+        if typ != 'OK':
+            return typ, data
         if len(data) != 1:
             raise self.error('GETSCRIPT returned more than one string/script')
         # todo: decode data?
         return typ, data[0][0]
-    
 
     def putscript(self, scriptname, scriptdata):
         """Put a script onto the server.
@@ -575,7 +572,7 @@ class MANAGESIEVE:
         :param str scriptname: name of script to be retrieved
         :param str scriptdata: script content
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         # command-putscript     = "PUTSCRIPT" SP sieve-name SP string CRLF
         # response-putscript    = response-oknobye
@@ -589,32 +586,30 @@ class MANAGESIEVE:
 
         :param str scriptname: name of script to be deleted
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         # command-deletescript  = "DELETESCRIPT" SP sieve-name CRLF
         # response-deletescript = response-oknobye
         return self._simple_command(b'DELETESCRIPT', sieve_name(scriptname))
 
-
     def setactive(self, scriptname):
         """Mark a script as the 'active' one.
 
         :param str scriptname: name of script to be marked active
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         # command-setactive     = "SETACTIVE" SP sieve-name CRLF
         # response-setactive    = response-oknobye
         return self._simple_command(b'SETACTIVE', sieve_name(scriptname))
 
-
     def havespace(self, scriptname, size):
         """Query the server for available space.
 
         :param str scriptname: name of script to XXX
         :param int size: XXX
 
-        :returns: response (:const:`OK`, :const:`NO`, :const:`BYE`)
+        :returns: response (``OK``, ``NO``, ``BYE``)
         """
         # command-havespace     = "HAVESPACE" SP sieve-name SP number CRLF
         # response-havespace    = response-oknobye
@@ -622,20 +617,19 @@ class MANAGESIEVE:
                                     sieve_name(scriptname),
                                     str(size).encode('ascii'))
 
-
     def capability(self):
         """
         Issue a CAPABILITY command and return the result.
-        
-        As a side-effect, on succes these attributes are (re)set:
 
-        - :attr:`capabilities` (list of strings)
-        - :attr:`loginmechs` (list of strings)
-        - :attr:`implementation` (string)
-        - :attr:`supports_tls` (boolean)
+        As a side-effect, on success these attributes are (re)set:
+
+        - :code:`capabilities` (list of strings)
+        - :code:`loginmechs` (list of strings)
+        - :code:`implementation` (string)
+        - :code:`supports_tls` (boolean)
 
         :returns: tuple(response, capabilities) --
-                  If `response` is :const:`OK`, `capabilities` is a list
+                  If `response` is ``OK``, `capabilities` is a list
                   of strings.
         """
         # command-capability    = "CAPABILITY" CRLF
@@ -645,8 +639,7 @@ class MANAGESIEVE:
             self._parse_capabilities(data)
         return typ, data
 
-
-    def starttls(self, keyfile=None, certfile=None):
+    def starttls(self, keyfile=None, certfile=None, verify=True):
         """Puts the connection to the SIEVE server into TLS mode.
 
         If the server supports TLS, this will encrypt the rest of the SIEVE
@@ -659,9 +652,15 @@ class MANAGESIEVE:
         # response-starttls     = response-oknobye
         typ, data = self._command(b'STARTTLS')
         if typ == 'OK':
-            sslobj = ssl_wrap_socket(self.sock, keyfile, certfile)
+            sslcontext = ssl.create_default_context()
+            if not verify:
+                sslcontext.check_hostname = False
+                sslcontext.verify_mode = ssl.CERT_NONE
+            if certfile:
+                sslcontext.load_cert_chain(certfile, keyfile)
+            sslobj = sslcontext.wrap_socket(
+                self.sock, server_hostname=self.host)
             self.sock = SSLFakeSocket(self.sock, sslobj)
-            self.file = SSLFakeFile(sslobj)
             # MUST discard knowledge obtained from the server
             self.__clear_knowledge()
             # NB: We did send a BOGUS command here for buggy servers,
